from pydantic.v1 import BaseModel, Field

class AgentVote(BaseModel):
    task: str = Field(description="Task to be executed by the agent")
    agent_vote: str = Field(description="Agent that will execute the task")



import json
from typing import List, Optional
from pydantic.v1 import BaseModel, Field, Json, root_validator

from .process import Process
from .agent import Agent
from .task import Task
from .tools.agent_tools import AgentTools

class Crew(BaseModel):
	"""
	Class that represents a group of agents, how they should work together and
	their tasks.
	"""
	config: Optional[Json] = Field(description="Configuration of the crew.")
	tasks: Optional[List[Task]] = Field(description="List of tasks")
	agents: Optional[List[Agent]] = Field(description="List of agents in this crew.")
	process: Process = Field(
		description="Process that the crew will follow.",
		default=Process.sequential
	)
	verbose: bool = Field(
		description="Verbose mode for the Agent Execution",
		default=False
	)

	@root_validator(pre=True)
	def check_config(_cls, values):
		if (
			not values.get('config') 
			and (
				not values.get('agents') and not values.get('tasks')
			)
		):
			raise ValueError('Either agents and task need to be set or config.')
		
		if values.get('config'):
			config = json.loads(values.get('config'))
			if not config.get('agents') or not config.get('tasks'):
				raise ValueError('Config should have agents and tasks.')
			
			values['agents'] = [Agent(**agent) for agent in config['agents']]
			
			tasks = []
			for task in config['tasks']:
				task_agent = [agt for agt in values['agents'] if agt.role == task['agent']][0]
				del task['agent']
				tasks.append(Task(**task, agent=task_agent))
			
			values['tasks'] = tasks
		return values

	def kickoff(self) -> str:
		"""
		Kickoff the crew to work on it's tasks.
			Returns:
				output (List[str]): Output of the crew for each task.
		"""
		if self.process == Process.sequential:
			return self.__sequential_loop()

	def __sequential_loop(self) -> str:
		"""
		Loop that executes the sequential process.
			Returns:
				output (str): Output of the crew.
		"""
		task_outcome = None
		for task in self.tasks:
			# Add delegation tools to the task if the agent allows it
			if task.agent.allow_delegation:
				tools = AgentTools(agents=self.agents).tools()
				task.tools += tools

			self.__log(f"\nWorking Agent: {task.agent.role}")
			self.__log(f"Starting Task: {task.description} ...")

			task_outcome = task.execute(task_outcome)

			self.__log(f"Task output: {task_outcome}")
		
		return task_outcome
	
	def __log(self, message):
		if self.verbose:
			print(message)


from typing import List, Any
from pydantic.v1 import BaseModel, Field
from textwrap import dedent
from langchain.tools import Tool

from ..agent import Agent

class AgentTools(BaseModel):
	"""Tools for generic agent."""
	agents: List[Agent] = Field(description="List of agents in this crew.")

	def tools(self):
		return [
			Tool.from_function(
				func=self.delegate_work,
				name="Delegate Work to Co-Worker",
				description=dedent(f"""Useful to delegate a specific task to one of the 
				following co-workers: [{', '.join([agent.role for agent in self.agents])}].
				The input to this tool should be a pipe (|) separated text of length
				three, representing the role you want to delegate it to,  the task and 
				information necessary. For example, `coworker|task|information`.
				""")
			),
			Tool.from_function(
				func=self.ask_question,
				name="Ask Question to Co-Worker",
				description=dedent(f"""Useful to ask a question, opinion or take from on
				of the following co-workers: [{', '.join([agent.role for agent in self.agents])}]. 
				The input to this tool should be a pipe (|) separated text of length 
				three, representing the role you want to ask it to,  the question and 
				information necessary. For example, `coworker|question|information`.
				""")
			),
		]

	def delegate_work(self, command):
		"""Useful to delegate a specific task to a coworker."""
		return self.__execute(command)

	def ask_question(self, command):
		"""Useful to ask a question, opinion or take from a coworker."""
		return self.__execute(command)

	def __execute(self, command):
		"""Execute the command."""
		agent, task, information = command.split("|")
		if not agent or not task or not information:
			return "Error executing tool. Missing 3 pipe (|) separated values."
		
		agent = [available_agent for available_agent in self.agents if available_agent.role == agent]
		
		if len(agent) == 0:
			return "Error executing tool. Co-worker not found, double check the co-worker."

		agent = agent[0]
		result = agent.execute_task(task, information)	
		return result


from typing import List, Optional
from pydantic.v1 import BaseModel, Field, root_validator

from langchain.tools import Tool

from .agent import Agent

class Task(BaseModel):
	"""
	Class that represent a task to be executed.
	"""
	
	description: str = Field(description="Description of the actual task.")
	agent: Optional[Agent] = Field(
		description="Agent responsible for the task.",
		default=None
	)
	tools: Optional[List[Tool]] = Field(
		description="Tools the agent are limited to use for this task.",
		default=[]
	)

	@root_validator(pre=False)
	def _set_tools(_cls, values):
		if (values.get('agent')) and not (values.get('tools')):
			values['tools'] = values.get('agent').tools
		return values

	def execute(self, context: str = None) -> str:
		"""
		Execute the task.
			Returns:
				output (str): Output of the task.
		"""
		if self.agent:
			return self.agent.execute_task(
				task = self.description, 
				context = context,
				tools = self.tools
			)
		else:
			raise Exception(f"The task '{self.description}' has no agent assigned, therefore it can't be executed directly and should be executed in a Crew using a specific process that support that, either consensual or hierarchical.")


from enum import Enum

class Process(str, Enum):
	"""
	Class representing the different processes that can be used to tackle tasks
	"""
	sequential = 'sequential'
	# TODO: consensual = 'consensual'
	# TODO: hierarchical = 'hierarchical'



"""Generic agent."""

from typing import List, Any, Optional
from pydantic.v1 import BaseModel, Field, root_validator

from langchain.agents import AgentExecutor
from langchain.chat_models import ChatOpenAI as OpenAI
from langchain.tools.render import render_text_description
from langchain.agents.format_scratchpad import format_log_to_str
from langchain.agents.output_parsers import ReActSingleInputOutputParser
from langchain.memory import ConversationSummaryMemory

from .prompts import Prompts

class Agent(BaseModel):
	"""Generic agent implementation."""
	agent_executor: AgentExecutor = None
	role: str = Field(description="Role of the agent")
	goal: str = Field(description="Objective of the agent")
	backstory: str = Field(description="Backstory of the agent")
	llm: Optional[OpenAI] = Field(description="LLM that will run the agent")
	verbose: bool = Field(
		description="Verbose mode for the Agent Execution",
		default=False
	)
	allow_delegation: bool = Field(
		description="Allow delegation of tasks to agents",
		default=True
	)
	tools: List[Any] = Field(
		description="Tools at agents disposal",
		default=[]
	)

	@root_validator(pre=True)
	def check_llm(_cls, values):
		if not values.get('llm'):
			values['llm'] = OpenAI(
				temperature=0.7,
				model_name="gpt-4"
			)
		return values

	def __init__(self, **data):
		super().__init__(**data)
		execution_prompt = Prompts.TASK_EXECUTION_PROMPT.partial(
			goal=self.goal,
			role=self.role,
			backstory=self.backstory,
		)

		llm_with_bind = self.llm.bind(stop=["\nObservation"])
		inner_agent = {
			"input": lambda x: x["input"],
			"tools": lambda x: x["tools"],
			"tool_names": lambda x: x["tool_names"],
			"chat_history": lambda x: x["chat_history"],
			"agent_scratchpad": lambda x: format_log_to_str(x['intermediate_steps']),
		} | execution_prompt | llm_with_bind | ReActSingleInputOutputParser()

		summary_memory = ConversationSummaryMemory(
			llm=self.llm,
			memory_key='chat_history',
			input_key="input"
		)

		self.agent_executor = AgentExecutor(
			agent=inner_agent,
			tools=self.tools,
			memory=summary_memory,
			verbose=self.verbose,
			handle_parsing_errors=True,
		)

	def execute_task(self, task: str, context: str = None, tools: List[Any] = None) -> str:
		"""
		Execute a task with the agent.
			Parameters:
				task (str): Task to execute
			Returns:
				output (str): Output of the agent
		"""
		if context:
			task = "\n".join([
				task,
				"\nThis is the context you are working with:",
				context
			])

		tools = tools or self.tools
		self.agent_executor.tools = tools
		return self.agent_executor.invoke({
			"input": task,
			"tool_names": self.__tools_names(tools),
			"tools": render_text_description(tools),
		})['output']

	def __tools_names(self, tools) -> str:
		return ", ".join([t.name for t in tools])


"""Prompts for generic agent."""

from textwrap import dedent
from typing import ClassVar
from pydantic.v1 import BaseModel
from langchain.prompts import PromptTemplate

class Prompts(BaseModel):
	"""Prompts for generic agent."""

	TASK_SLICE: ClassVar[str] = dedent("""\
		Begin! This is VERY important to you, your job depends on it!

		Current Task: {input}
		{agent_scratchpad}
	""")

	MEMORY_SLICE: ClassVar[str] = dedent("""\
		This is the summary of your work so far:
        {chat_history}
	""")

	ROLE_PLAYING_SLICE: ClassVar[str] = dedent("""\
		You are {role}.
		{backstory}

		Your personal goal is: {goal}
	""")

	TOOLS_SLICE: ClassVar[str] = dedent("""\
		TOOLS:
		------

		You have access to the following tools:

		{tools}

		To use a tool, please use the following format:

		```
		Thought: Do I need to use a tool? Yes
		Action: the action to take, should be one of [{tool_names}]
		Action Input: the input to the action
		Observation: the result of the action
		```

		When you have a response for your task, or if you do not need to use a tool, you MUST use the format:

		```
		Thought: Do I need to use a tool? No
		Final Answer: [your response here]
		```
	""")

	VOTING_SLICE: ClassVar[str] = dedent("""\
		You are working on a crew with your co-workers and need to decide who will execute the task.

		These are tyour format instructions:
		{format_instructions}

		These are your co-workers and their roles:
		{coworkers}
	""")

	TASK_EXECUTION_PROMPT: ClassVar[str] = PromptTemplate.from_template(
		ROLE_PLAYING_SLICE + TOOLS_SLICE + MEMORY_SLICE + TASK_SLICE
	)
	
	CONSENSUNS_VOTING_PROMPT: ClassVar[str] = PromptTemplate.from_template(
		ROLE_PLAYING_SLICE + VOTING_SLICE + TASK_SLICE
	)